#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


import abc
import asyncio
import time
from dataclasses import dataclass
from http import HTTPStatus
from typing import Awaitable, Callable, Optional

import aiohttp

from meatie.aio import (
    Client,
    Context,
    Duration,
    RateLimitExceeded,
    Response,
)
from meatie.aio.internal import EndpointDescriptor
from meatie.internal.types import PT, T


@dataclass()
class RetryContext:
    attempt_number: int
    started_at: float
    error: Optional[Exception]
    response: Optional[Response]


class _BaseCondition:
    @abc.abstractmethod
    def __call__(self, ctx: RetryContext) -> bool:
        pass

    def __and__(self, other: "_BaseCondition") -> "_BaseCondition":
        return _AndCondition(self, other)

    def __or__(self, other: "_BaseCondition") -> "_BaseCondition":
        return _OrCondition(self, other)


class _AndCondition(_BaseCondition):
    def __init__(self, left: _BaseCondition, right: _BaseCondition) -> None:
        self._left = left
        self._right = right

    def __call__(self, ctx: RetryContext) -> bool:
        return self._left(ctx) and self._right(ctx)


class _OrCondition(_BaseCondition):
    def __init__(self, left: _BaseCondition, right: _BaseCondition) -> None:
        self._left = left
        self._right = right

    def __call__(self, ctx: RetryContext) -> bool:
        return self._left(ctx) or self._right(ctx)


def _no_wait(_: RetryContext) -> float:
    return 0.0


class RetryOnStatusCode(_BaseCondition):
    def __init__(self, status: int) -> None:
        self.status = status

    def __call__(self, ctx: RetryContext) -> bool:
        if ctx.response is not None:
            return ctx.response.status == self.status

        if isinstance(ctx.error, aiohttp.ClientResponseError):
            return ctx.error.status == self.status

        return False


class RetryOnExceptionType(_BaseCondition):
    def __init__(self, exc_type: type[BaseException]) -> None:
        self.exc_type = exc_type

    def __call__(self, ctx: RetryContext) -> bool:
        return isinstance(ctx.error, self.exc_type)


RetryOnServerConnectionError = RetryOnExceptionType(aiohttp.ServerConnectionError)
RetryOnTooManyRequestsStatus = RetryOnStatusCode(HTTPStatus.TOO_MANY_REQUESTS)
NoWait = _no_wait


class WaitExponential(_BaseCondition):
    def __init__(self, multiplier: int = 2, lower_bound: int = 1, upper_bound: int = 3600) -> None:
        self._multiplier = multiplier
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound

    def __call__(self, ctx: RetryContext) -> bool:
        return max(min(self._multiplier**ctx.attempt_number, self._upper_bound), self._lower_bound)


def _never_stop(ctx: RetryContext) -> bool:
    return False


class StopAfter:
    def __init__(self, attempts: int) -> None:
        self.attempts = attempts

    def __call__(self, ctx: RetryContext) -> bool:
        return ctx.attempt_number > self.attempts


NeverStop = _never_stop

StopCondition = Callable[[RetryContext], bool]
RetryCondition = Callable[[RetryContext], bool]
WaitFunc = Callable[[RetryContext], float]


class RetryOption:
    __PRIORITY = 75

    def __init__(
        self,
        retry: RetryCondition = (RetryOnTooManyRequestsStatus | RetryOnServerConnectionError),
        wait: WaitFunc = NoWait,
        stop: StopCondition = NeverStop,
        sleep_func: Callable[[Duration], Awaitable[None]] = asyncio.sleep,
    ) -> None:
        self.__retry = retry
        self.__wait = wait
        self.__stop = stop
        self.__sleep_func = sleep_func

    def __call__(self, descriptor: EndpointDescriptor[PT, T]) -> None:
        descriptor.register_operator(RetryOption.__PRIORITY, self.__operator)

    async def __operator(self, operation_ctx: Context[Client, T]) -> T:
        retry_ctx = RetryContext(
            attempt_number=1, started_at=time.monotonic(), error=None, response=None
        )
        last_result: Optional[T] = None
        stopped = False
        while not stopped:
            if retry_ctx.attempt_number > 1:
                wait_time = self.__wait(retry_ctx)
                if wait_time > 0.0:
                    await self.__sleep_func(wait_time)

            try:
                last_result = await operation_ctx.proceed()
                retry_ctx.response = operation_ctx.response
            except Exception as exc:
                retry_ctx.error = exc

            if not self.__retry(retry_ctx):
                break

            retry_ctx.attempt_number += 1
            stopped = self.__stop(retry_ctx)

        if not stopped and last_result is not None:
            return last_result

        if retry_ctx.error is not None:
            raise retry_ctx.error
        raise RateLimitExceeded()



#  Copyright 2024 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import importlib.metadata

__version__ = importlib.metadata.version("meatie")



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

from meatie.aio import ParseResponseError, Response

from .types import TypeAdapter


class _StringAdapter:
    @staticmethod
    async def from_response(response: Response) -> str:
        try:
            return await response.text()
        except ValueError as exc:
            raise ParseResponseError(response.status, None) from exc

    @staticmethod
    def to_json(value: str) -> str:  # pragma: no cover
        return value


StringAdapter: TypeAdapter[str] = _StringAdapter()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


from typing import Any, Awaitable, Callable, Optional, Protocol, TypeVar

from typing_extensions import Self

from meatie.internal.types import T_Out

from . import Request, Response


class Client(Protocol):
    async def make_request(self, request: Request) -> Response:
        ...

    async def __aenter__(self) -> Self:
        ...

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Any,
    ) -> None:
        ...


ClientType = TypeVar("ClientType", bound=Client, covariant=True)


class Context(Protocol[ClientType, T_Out]):
    @property
    def client(self) -> ClientType:
        ...

    @property
    def request(self) -> Request:
        ...

    @property
    def response(self) -> Optional[Response]:
        ...

    async def proceed(self) -> T_Out:
        ...


Operator = Callable[[Context[ClientType, T_Out]], Awaitable[T_Out]]



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import re
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    Optional,
    Protocol,
    Union,
    get_args,
    overload,
    runtime_checkable,
)

from aiohttp import ClientResponse
from typing_extensions import Self

from meatie.aio import (
    Client,
    Method,
    Operator,
    Request,
)
from meatie.internal.types import PT

from .request_template import RequestTemplate, ResponseBodyT


class EndpointDescriptor(Generic[PT, ResponseBodyT]):
    __slots__ = (
        "template",
        "__content_decoder",
        "__operator_by_priority",
    )

    def __init__(self, template: RequestTemplate[Any, ResponseBodyT]) -> None:
        self.template = template
        self.__operator_by_priority: dict[int, Operator[Client, ResponseBodyT]] = {}

    def __set_name__(self, owner: type[object], name: str) -> None:
        if self.template.method is not None:
            return

        method_opt = _get_method_opt(name)
        if method_opt is None:
            # Python interactions with class descriptors doesn't allow to return an error message to the user,
            # we assume the GET method
            method_opt = "GET"

        self.template.method = method_opt

    def register_operator(self, priority: int, operator: Operator[Client, ResponseBodyT]) -> None:
        self.__operator_by_priority[priority] = operator

    @overload
    def __get__(self, instance: None, owner: None) -> Self:
        ...

    @overload
    def __get__(
        self, instance: Client, owner: type[object]
    ) -> Callable[PT, Awaitable[ResponseBodyT]]:
        ...

    def __get__(
        self, instance: Optional[Client], owner: Optional[type[object]] = None
    ) -> Union[Self, Callable[PT, Awaitable[ResponseBodyT]]]:
        if instance is None or owner is None:
            return self

        priority_operator_pair = list(self.__operator_by_priority.items())
        priority_operator_pair.sort()
        operators = [operator for _, operator in priority_operator_pair]

        return _BoundEndpointDescriptor(  # type: ignore[return-value]
            instance, operators, self.template
        )


_method_pattern_pairs = [
    (method, re.compile("^" + method, re.IGNORECASE)) for method in get_args(Method)
]


def _get_method_opt(name: str) -> Optional[Method]:
    for method, pattern in _method_pattern_pairs:
        if re.match(pattern, name):
            return method
    return None


@runtime_checkable
class EndpointOption(Protocol):
    def __call__(self, endpoint: EndpointDescriptor[PT, ResponseBodyT]) -> None:
        ...


class _Context(Generic[ResponseBodyT]):
    def __init__(
        self, client: Client, operators: list[Operator[Client, ResponseBodyT]], request: Request
    ) -> None:
        self.client = client
        self.__operators = operators
        self.__next_step = 0

        self.request = request
        self.response: Optional[ClientResponse] = None

    async def proceed(self) -> ResponseBodyT:
        if self.__next_step >= len(self.__operators):
            raise RuntimeError("No more step to process request")

        current_step = self.__next_step
        self.__next_step += 1
        current_operator = self.__operators[current_step]
        try:
            result = await current_operator(self)
            return result
        finally:
            self.__next_step = current_step


class _BoundEndpointDescriptor(Generic[PT, ResponseBodyT]):
    def __init__(
        self,
        instance: Client,
        operators: Iterable[Operator[Client, ResponseBodyT]],
        template: RequestTemplate[Any, ResponseBodyT],
    ) -> None:
        self.__instance = instance
        self.__operators = list(operators)
        self.__operators.append(self.__make_request)  # type: ignore[arg-type]
        self.__template = template

    async def __call__(self, *args: PT.args, **kwargs: PT.kwargs) -> ResponseBodyT:
        request = self.__template.build_request(*args, **kwargs)
        context = _Context[ResponseBodyT](self.__instance, self.__operators, request)
        return await context.proceed()

    async def __make_request(self, context: _Context[ResponseBodyT]) -> ResponseBodyT:
        response = await self.__instance.make_request(context.request)
        context.response = response
        return await self.__template.response_decoder.from_response(response)



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import inspect
import re
from enum import Enum
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Optional,
    TypeVar,
    get_args,
    get_type_hints,
)

from typing_extensions import Awaitable, Self

from meatie.aio.internal import JsonAdapter, TypeAdapter, get_adapter
from meatie.internal.http import Method, Request
from meatie.internal.types import PT


class Kind(Enum):
    UNKNOWN = 0
    PATH = 1
    QUERY = 2
    BODY = 3


class ApiRef:
    __slots__ = ("name",)

    def __init__(self, name: str) -> None:
        self.name = name

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ApiRef):
            return self.name == other.name
        return False

    @classmethod
    def from_signature(cls, parameter: inspect.Parameter) -> Self:
        for arg in get_args(parameter.annotation):
            if isinstance(arg, cls):
                return arg
        return cls(name=parameter.name)


class Parameter:
    __slots__ = ("kind", "name", "api_ref", "default_value")

    def __init__(self, kind: Kind, name: str, api_ref: str, default_value: Any = None) -> None:
        self.kind = kind
        self.name = name
        self.api_ref = api_ref
        self.default_value = default_value

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Parameter):
            return (
                self.name == other.name
                and self.kind == other.kind
                and self.api_ref == other.api_ref
                and self.default_value == other.default_value
            )
        return False


_param_pattern = re.compile(r"{(?P<name>[a-zA-Z_]\w*?)}")


class PathTemplate:
    __slots__ = ("template", "parameters")

    def __init__(self, template: str, parameters: list[str]) -> None:
        self.template = template
        self.parameters = parameters

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, PathTemplate):
            return self.template == other.template and self.parameters == other.parameters
        if isinstance(other, str):
            return self.template == other
        return False

    def __hash__(self) -> int:
        return hash(self.template)

    def __contains__(self, item: str) -> bool:
        return item in self.parameters

    def __str__(self) -> str:
        return self.template

    def format(self, **kwargs: dict[str, Any]) -> str:
        return self.template.format(**kwargs)

    @classmethod
    def from_string(cls, template: str) -> Self:
        parameters = [match.group("name") for match in _param_pattern.finditer(template)]
        return cls(template, parameters)


RequestBodyT = TypeVar("RequestBodyT")
ResponseBodyT = TypeVar("ResponseBodyT")


class RequestTemplate(Generic[RequestBodyT, ResponseBodyT]):
    __slots__ = (
        "method",
        "template",
        "params",
        "request_encoder",
        "response_decoder",
        "__param_by_name",
    )

    def __init__(
        self,
        template: PathTemplate,
        params: list[Parameter],
        request_encoder: TypeAdapter[RequestBodyT],
        response_decoder: TypeAdapter[ResponseBodyT],
        method: Optional[Method],
    ) -> None:
        self.method = method
        self.template = template
        self.params = params
        self.request_encoder = request_encoder
        self.response_decoder = response_decoder

        self.__param_by_name: dict[str, Parameter] = {}
        for param in self.params:
            self.__param_by_name[param.name] = param

    def build_request(self, *args: Any, **kwargs: Any) -> Request:
        if self.method is None:
            raise RuntimeError("'method' is None")

        if len(args) > len(self.params):
            raise RuntimeError("Too many arguments passed to the function.")

        value_by_param = {}
        for index, value in enumerate(args):
            param = self.params[index]
            value_by_param[param] = value

        for param in self.params:
            if param.default_value is None or param in value_by_param:
                continue
            value_by_param[param] = param.default_value

        for name, value in kwargs.items():
            param_opt = self.__param_by_name.get(name)
            if param_opt is None:
                raise ValueError(f"Parameter '{name}' is not mentioned in the endpoint definition.")

            if value is None:
                if param_opt in value_by_param:
                    del value_by_param[param_opt]
            else:
                value_by_param[param_opt] = value

        path_kwargs = {}
        query_kwargs = {}
        body_value: Any = None
        for param, value in value_by_param.items():
            if param.kind == Kind.PATH:
                path_kwargs[param.api_ref] = value
                continue

            if param.kind == Kind.QUERY:
                query_kwargs[param.api_ref] = value
                continue

            if param.kind == Kind.BODY:
                body_value = value
                continue

            raise NotImplementedError(f"Kind {param.kind} is not supported")  # pragma: no cover

        if body_value is not None:
            body_json = self.request_encoder.to_json(body_value)
        else:
            body_json = None

        path = self.template.format(**path_kwargs)
        return Request(
            method=self.method, path=path, query_params=query_kwargs, headers={}, json=body_json
        )

    @classmethod
    def from_callable(
        cls,
        func: Callable[PT, Awaitable[ResponseBodyT]],
        template: PathTemplate,
        method: Optional[Method],
    ) -> Self:
        signature = inspect.signature(func)
        type_by_param_name = get_type_hints(func)

        parameters = []
        request_encoder: TypeAdapter[Any] = JsonAdapter
        response_decoder: TypeAdapter[ResponseBodyT] = JsonAdapter
        for param_name in type_by_param_name:
            param_type = type_by_param_name[param_name]
            if param_name == "return":
                response_decoder = get_adapter(param_type)
                continue

            sig_param = signature.parameters[param_name]
            api_ref = ApiRef.from_signature(sig_param)
            kind = Kind.QUERY
            if api_ref.name == "body":
                kind = Kind.BODY
                request_encoder = get_adapter(param_type)
            elif api_ref.name in template:
                kind = Kind.PATH

            default_value = None
            if sig_param.default is not inspect.Parameter.empty:
                default_value = sig_param.default
            parameter = Parameter(kind, param_name, api_ref.name, default_value)
            parameters.append(parameter)

        return cls.validate_object(
            template, parameters, signature, request_encoder, response_decoder, method
        )

    @classmethod
    def validate_object(
        cls,
        template: PathTemplate,
        params: Iterable[Parameter],
        signature: inspect.Signature,
        request_encoder: TypeAdapter[RequestBodyT],
        response_decoder: TypeAdapter[ResponseBodyT],
        method: Optional[Method],
    ) -> Self:
        template_str = str(template)
        if template_str == "":
            raise ValueError("'path' is empty")

        if not template_str.startswith("/"):
            raise ValueError("'path' must start with '/'")

        visited_api_refs = set()
        for param in params:
            if param.name == "self":
                raise ValueError("Parameter 'self' is forbidden")

            if param.api_ref in visited_api_refs:
                raise ValueError(f"Multiple parameters have name '{param.api_ref}'")

            if param.kind == Kind.PATH and param.api_ref not in template:
                raise ValueError(
                    f"Parameter '{param.api_ref}' is not present in the path '{template}'"
                )

            visited_api_refs.add(param.api_ref)

        missing_api_refs = set(template.parameters) - visited_api_refs
        for api_ref in missing_api_refs:
            raise ValueError(
                f"Parameter '{api_ref}' is not present in the method signature '{signature}'"
            )

        return cls(template, list(params), request_encoder, response_decoder, method)



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


import json
from typing import Any, Callable, Optional, Protocol


class Response(Protocol):
    @property
    def status(self) -> int:
        ...

    async def read(self) -> bytes:
        ...

    async def text(self, encoding: Optional[str] = None) -> str:
        ...

    async def json(
        self, *, encoding: Optional[str] = None, loads: Callable[[str], Any] = json.loads
    ) -> dict[str, Any]:
        ...



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

# isort:skip_file
from .adapter import TypeAdapter, JsonAdapter, get_adapter
from .request_template import RequestTemplate, ApiRef, PathTemplate
from .endpoint_descriptor import EndpointDescriptor, EndpointOption

__all__ = [
    "TypeAdapter",
    "JsonAdapter",
    "get_adapter",
    "RequestTemplate",
    "ApiRef",
    "PathTemplate",
    "EndpointDescriptor",
    "EndpointOption",
]



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

# isort:skip_file

from meatie.internal.number import INF
from meatie.internal.http import Method, Request
from meatie.internal.time_ import Time, Duration, DAY, HOUR, MINUTE
from meatie.internal.error import MeatieError, RateLimitExceeded, ParseResponseError
from meatie.internal.limit import Rate, Limiter, Reservation, Tokens
from meatie.internal.cache import CacheStore
from .response import Response
from .client import Client
from .types import Context, Operator
from .internal import ApiRef, EndpointDescriptor
from .option import (
    Cache,
    Private,
    Limit,
    Retry,
    WaitExponential,
    NoWait,
    StopAfter,
    RetryOnStatusCode,
    RetryOnExceptionType,
    RetryOnServerConnectionError,
    RetryOnTooManyRequestsStatus,
)
from .endpoint import endpoint

__all__ = [
    "INF",
    "Method",
    "Request",
    "Time",
    "Duration",
    "MINUTE",
    "HOUR",
    "DAY",
    "Response",
    "MeatieError",
    "RateLimitExceeded",
    "ParseResponseError",
    "Rate",
    "Limiter",
    "Reservation",
    "Tokens",
    "ApiRef",
    "EndpointDescriptor",
    "CacheStore",
    "Cache",
    "Private",
    "Limit",
    "Retry",
    "WaitExponential",
    "NoWait",
    "StopAfter",
    "RetryOnStatusCode",
    "RetryOnExceptionType",
    "RetryOnServerConnectionError",
    "RetryOnTooManyRequestsStatus",
    "Client",
    "Context",
    "Operator",
    "endpoint",
]



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

from dataclasses import dataclass
from typing import Any, Literal, Optional, Union

Method = Literal["GET", "HEAD", "POST", "PUT", "DELETE", "OPTIONS", "TRACE", "PATCH"]


@dataclass()
class Request:
    method: Method
    path: str
    query_params: dict[str, Union[str, int]]
    headers: dict[str, Union[str, bytes]]
    data: Optional[bytes] = None
    json: Any = None



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


from meatie.aio import Client, Context
from meatie.aio.internal import EndpointDescriptor
from meatie.internal.types import PT, T


class PrivateOption:
    __PRIORITY = 50

    def __call__(self, descriptor: EndpointDescriptor[PT, T]) -> None:
        descriptor.register_operator(PrivateOption.__PRIORITY, PrivateOption.__operator)

    @staticmethod
    async def __operator(ctx: Context[Client, T]) -> T:
        await ctx.client.authenticate(ctx.request)
        return await ctx.proceed()


Instance = PrivateOption()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


from meatie.aio import Response

from .types import TypeAdapter


class _BytesAdapter:
    @staticmethod
    async def from_response(response: Response) -> bytes:
        return await response.read()

    @staticmethod
    def to_json(value: bytes) -> bytes:
        return value


BytesAdapter: TypeAdapter[bytes] = _BytesAdapter()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from typing import Optional


class MeatieError(Exception):
    ...


class RateLimitExceeded(MeatieError):
    ...


class ResponseError(MeatieError):
    __slots__ = ("status", "message")

    def __init__(self, status: int, message: Optional[str]) -> None:
        """
        :param status: HTTP status code
        :param message: Message of the response
        """

        self.status = status
        self.message = message


class ParseResponseError(ResponseError):
    ...



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

from collections.abc import Callable
from typing import (
    Any,
    Awaitable,
    Optional,
)

from meatie.aio.internal import (
    EndpointDescriptor,
    EndpointOption,
    PathTemplate,
    RequestTemplate,
)
from meatie.internal.http import Method
from meatie.internal.types import PT, T


def endpoint(
    path: str,
    *args: Any,
    method: Optional[Method] = None,
) -> Callable[[Callable[PT, Awaitable[T]]], Callable[PT, Awaitable[T]]]:
    def class_descriptor(
        func: Callable[PT, Awaitable[T]],
    ) -> Callable[PT, Awaitable[T]]:
        path_template = PathTemplate.from_string(path)
        request_template: RequestTemplate[Any, T] = RequestTemplate.from_callable(
            func, path_template, method
        )
        descriptor = EndpointDescriptor[PT, T](request_template)

        for option in args:
            if isinstance(option, EndpointOption):
                option(descriptor)

        return descriptor  # type: ignore[return-value]

    return class_descriptor



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

from .cache import CacheOption as Cache
from .limit import LimitOption as Limit
from .private import Instance as Private
from .retry import (
    NeverStop,
    NoWait,
    RetryOnExceptionType,
    RetryOnServerConnectionError,
    RetryOnStatusCode,
    RetryOnTooManyRequestsStatus,
    StopAfter,
    WaitExponential,
)
from .retry import RetryOption as Retry

__all__ = [
    "Cache",
    "Limit",
    "Retry",
    "NoWait",
    "WaitExponential",
    "NeverStop",
    "StopAfter",
    "RetryOnStatusCode",
    "RetryOnTooManyRequestsStatus",
    "RetryOnExceptionType",
    "RetryOnServerConnectionError",
    "Private",
]



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


import urllib.parse
from typing import Any, Optional

from aiohttp import ClientResponse, ClientSession
from typing_extensions import Self

from . import INF, CacheStore, Limiter, Rate, Request


class Client:
    shared_cache: CacheStore

    def __init__(
        self,
        session: ClientSession,
        session_params: Optional[dict[str, Any]] = None,
        local_cache: Optional[CacheStore] = None,
        limiter: Optional[Limiter] = None,
    ):
        self.session = session
        self.local_cache = local_cache if local_cache is not None else CacheStore()
        self.limiter = limiter if limiter is not None else Limiter(Rate.max, INF)
        self.session_params = session_params if session_params else {}

    def __init_subclass__(cls, **kwargs: Any) -> None:
        cls.shared_cache = CacheStore()

    async def make_request(self, request: Request) -> ClientResponse:
        kwargs: dict[str, Any] = self.session_params.copy()

        if request.data is not None:
            kwargs["data"] = request.data

        if request.json is not None:
            kwargs["json"] = request.json

        if request.headers:
            kwargs["headers"] = request.headers

        if request.query_params:
            path = request.path + "?" + urllib.parse.urlencode(request.query_params)
        else:
            path = request.path

        response = await self.session.request(request.method, path, **kwargs)
        return response

    async def authenticate(self, request: Request) -> None:
        pass

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Any,
    ) -> None:
        await self.close()

    async def close(self) -> None:
        await self.session.close()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from typing import TypeVar

from typing_extensions import ParamSpec

VT = TypeVar("VT")
T = TypeVar("T")
T_In = TypeVar("T_In", contravariant=True)
T_Out = TypeVar("T_Out", covariant=True)
PT = ParamSpec("PT")



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from json import JSONDecodeError
from typing import Any

from aiohttp import ContentTypeError

from meatie.aio import ParseResponseError, Response


class _JsonAdapter:
    @staticmethod
    async def from_response(response: Response) -> Any:
        try:
            return await response.json()
        except ContentTypeError as exc:
            message = await response.text()
            raise ParseResponseError(response.status, message) from exc
        except JSONDecodeError as exc:
            message = await response.text()
            raise ParseResponseError(response.status, message) from exc

    @staticmethod
    def to_json(value: Any) -> Any:
        return value


JsonAdapter = _JsonAdapter()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

import abc
import urllib.parse
from typing import Generic

from meatie.aio import (
    CacheStore,
    Client,
    Context,
    Duration,
    Operator,
)
from meatie.aio.internal import EndpointDescriptor
from meatie.internal.types import PT, T


class CacheOption:
    __PRIORITY = 25

    def __init__(self, ttl: Duration, shared: bool = False) -> None:
        self.ttl = ttl
        self.shared = shared

    def __call__(self, descriptor: EndpointDescriptor[PT, T]) -> None:
        operator: Operator[Client, T]
        if self.shared:
            operator = _SharedOperator[T](self.ttl)
        else:
            operator = _LocalOperator[T](self.ttl)

        descriptor.register_operator(CacheOption.__PRIORITY, operator)


class _Operator(Generic[T]):
    def __init__(self, ttl: Duration) -> None:
        self.ttl = ttl

    async def __call__(self, ctx: Context[Client, T]) -> T:
        key = ctx.request.path
        if ctx.request.query_params:
            key += "?" + urllib.parse.urlencode(ctx.request.query_params)

        storage = self._storage(ctx)
        value_opt = storage.load(key)
        if value_opt is not None:
            return value_opt

        value = await ctx.proceed()
        storage.store(key, value, self.ttl)
        return value

    @abc.abstractmethod
    def _storage(self, ctx: Context[Client, T]) -> CacheStore:
        ...


class _LocalOperator(_Operator[T]):
    def _storage(self, ctx: Context[Client, T]) -> CacheStore:
        return ctx.client.local_cache


class _SharedOperator(_Operator[T]):
    def _storage(self, ctx: Context[Client, T]) -> CacheStore:
        return ctx.client.shared_cache



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

from .duration import DAY, HOUR, MINUTE
from .types import Duration, Time

__all__ = ["Time", "Duration", "MINUTE", "HOUR", "DAY"]



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.

from meatie.aio import Response


class _NoneAdapter:
    @staticmethod
    async def from_response(_: Response) -> None:  # pragma: no cover
        return None

    @staticmethod
    def to_json(_: None) -> None:  # pragma: no cover
        return None


NoneAdapter = _NoneAdapter()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.


import asyncio
import time as sys_time
from typing import Awaitable, Callable

from meatie.aio import Client, Context
from meatie.aio.internal import EndpointDescriptor
from meatie.internal.limit import Tokens
from meatie.internal.types import PT, T


class LimitOption:
    __PRIORITY = 80

    def __init__(
        self, tokens: Tokens, sleep_func: Callable[[Tokens], Awaitable[None]] = asyncio.sleep
    ) -> None:
        self.tokens = tokens
        self.sleep_func = sleep_func

    def __call__(self, descriptor: EndpointDescriptor[PT, T]) -> None:
        if self.tokens > 0.0:
            descriptor.register_operator(LimitOption.__PRIORITY, self.__operator)

    async def __operator(self, ctx: Context[Client, T]) -> T:
        current_time = sys_time.monotonic()
        reservation = ctx.client.limiter.reserve_at(current_time, self.tokens)
        delay = reservation.ready_at - current_time
        if delay > 0:
            await self.sleep_func(delay)

        return await ctx.proceed()



#  Copyright 2023 The Meatie Authors. All rights reserved.
#  Use of this source code is governed by a BSD-style license that can be found in the LICENSE file.
from dataclasses import is_dataclass
from inspect import isclass
from typing import Any, Generic

from pydantic import BaseModel, ValidationError
from pydantic import TypeAdapter as pydantic_TypeAdapter
from typing_extensions import is_typeddict

from meatie.aio import Response
from meatie.internal.error import ParseResponseError
from meatie.internal.types import T

from . import JsonAdapter, TypeAdapter


class PydanticV2TypeAdapter(Generic[T]):
    def __init__(self, model_type: type[T]) -> None:
        self.adapter = pydantic_TypeAdapter(model_type)

    async def from_response(self, response: Response) -> T:
        json_model = await JsonAdapter.from_response(response)
        try:
            return self.adapter.validate_python(json_model)
        except ValidationError as exc:
            text = await response.text()
            raise ParseResponseError(response.status, text) from exc

    def to_json(self, value: T) -> Any:
        return self.adapter.dump_python(value, mode="json", by_alias=True)


class PydanticV2TypeAdapterFactory:
    @staticmethod
    def __call__(model_cls: type[T]) -> TypeAdapter[T]:
        return PydanticV2TypeAdapter(model_cls)

    @staticmethod
    def is_model_type(value: type[Any]) -> bool:
        return isclass(value) and (
            issubclass(value, BaseModel) or is_dataclass(value) or is_typeddict(value)
        )
